@@ -1,5 +1,6 @@
 #!/usr/bin/perl
 
+use 5.014; # s///r
 use strict;
 use warnings;
 
@@ -22,6 +23,8 @@ use Getopt::Long;
 use YAML;
 use Data::Dump 'pp';
 
+use Image::ExifTool;
+
 my $CONFIG = "client.yaml";
 
 GetOptions(
@@ -449,6 +452,7 @@ sub cmd_msg
 package RoomTab {
    use base qw( Tickit::Console::Tab );
 
+   use List::Util qw( any );
    use POSIX qw( strftime );
 
    use Net::Async::Matrix::Utils qw( parse_formatted_message );
@@ -475,23 +479,9 @@ package RoomTab {
       $presence_table->add( 0, 1, Tickit::Widget::Static->new( text => "Since" ) );
       $presence_table->add( 0, 2, Tickit::Widget::Static->new( text => "Lvl" ) );
 
-      my $presence_float = $floatbox->add_float(
-         child => Tickit::Widget::Frame->new(
-            style => {
-               linetype => "none",
-               linetype_left => "single",
-
-               frame_fg => "white", frame_bg => "purple",
-            },
-            child => my $vbox = Tickit::Widget::VBox->new,
-         ),
-
-         top => 0, bottom => -1, right => -1,
-         left => -44,
-
-         # Initially hidden
-         hidden => 1,
-      );
+      # Create an abstract widget tree during initial loading to avoid the
+      # O(n^2) overhead of resizing the gridbox after -every- user is added.
+      my $vbox = Tickit::Widget::VBox->new;
 
       $vbox->add(
          Tickit::Widget::ScrollBox->new(
@@ -506,6 +496,7 @@ package RoomTab {
          my $presence_summary = Tickit::Widget::Static->new( text => "" )
       );
 
+      my $presence_float;
       my $visible = 0;
       $self->bind_key( 'F2' => sub {
          $visible ? ( $presence_float->hide, $visible = 0 )
@@ -527,6 +518,25 @@ package RoomTab {
             );
 
             $room->paginate_messages( limit => 150 );
+
+            # Only now should we add the presence table to the floatbox
+            $presence_float = $floatbox->add_float(
+               child => Tickit::Widget::Frame->new(
+                  style => {
+                     linetype => "none",
+                     linetype_left => "single",
+
+                     frame_fg => "white", frame_bg => "purple",
+                  },
+                  child => $vbox,
+               ),
+
+               top => 0, bottom => -1, right => -1,
+               left => -44,
+
+               # Initially hidden
+               hidden => 1,
+            );
          },
 
          on_message => sub {
@@ -814,6 +824,18 @@ package RoomTab {
          },
       );
 
+      my $content_url;
+      if( $content->{url} ) {
+         my $uri = URI->new( $content->{url} );
+         if( $uri->scheme eq "mxc" ) {
+            $content_url = ( $SSL ? "https" : "http" ) .
+                           "://$SERVER/_matrix/media/v1/download/" . $uri->authority . $uri->path;
+         }
+         else {
+            $content_url = "$uri";
+         }
+      }
+
       if( $msgtype eq "m.text" ) {
          return $s
             ->append_tagged( "<", fg => "magenta" )
@@ -828,17 +850,48 @@ package RoomTab {
             ->append_tagged( " " )
             ->append       ( $body );
       }
-      elsif( $msgtype eq "m.image" ) {
-         my $url  = $content->{url};
+      elsif( $msgtype eq "m.notice" ) {
+         return $s
+            ->append_tagged( "--", fg => "red" )
+            ->append( format_displayname( $member ) )
+            ->append_tagged( "-- ", fg => "red" )
+            ->append       ( $body );
+      }
+      # Handle all the four attachment-style messages similarly
+      elsif( any { $msgtype eq $_ } qw( m.image m.audio m.video m.file ) ) {
          my $info = $content->{info} // $content->{body}; # cope with older message format
 
-         return $s
-            ->append_tagged( "[" )
-            ->append( format_displayname( $member ) )
-            ->append_tagged( "] " )
-            ->append_tagged( "image ", fg => "yellow" )
-            ->append_tagged( "($info->{w}x$info->{h}) ", fg => "grey" )
-            ->append_tagged( $url, fg => "hi-blue", u => 1 );
+         $s->append_tagged( "[" )
+           ->append( format_displayname( $member ) )
+           ->append_tagged( "] " )
+           ->append_tagged( $msgtype =~ s/^m\.//r, fg => "yellow" );
+
+         if( defined $info->{mimetype} ) {
+            $s->append_tagged( "; $info->{mimetype}", fg => "grey" );
+         }
+         if( defined( my $bytes = $info->{size} ) ) {
+            $s->append_tagged( "; " . format_bytes( $bytes ), fg => "grey" );
+         }
+
+         if( $msgtype eq "m.image" ) {
+             $s->append_tagged( " ($info->{w}x$info->{h})", fg => "grey" );
+         }
+         elsif( $msgtype eq "m.audio" ) {
+             $s->append_tagged( " (" . format_msec( $info->{duration} ) . ")", fg => "grey" );
+         }
+         elsif( $msgtype eq "m.video" ) {
+             $s->append_tagged( " ($info->{w}x$info->{h}, " . format_msec( $info->{duration} ) . ")", fg => "grey" );
+         }
+
+         $s->append_tagged( " " )
+           ->append_tagged( $content_url, fg => "hi-blue", u => 1 );
+
+         # filename comes from content, not info
+         if( defined $content->{filename} ) {
+            $s->append_tagged( " - $content->{filename}" );
+         }
+
+         return $s;
       }
       else {
          return $s
@@ -851,6 +904,25 @@ package RoomTab {
       }
    }
 
+   sub format_bytes
+   {
+      my ( $v ) = @_;
+      return sprintf "%d bytes", $v if $v < 1024; $v /= 1024;
+      return sprintf "%.1f KiB", $v if $v < 1024; $v /= 1024;
+      return sprintf "%.1f MiB", $v if $v < 1024; $v /= 1024;
+      return sprintf "%.1f GiB", $v if $v < 1024; $v /= 1024;
+      return sprintf "%.1f TiB", $v;
+   }
+
+   sub format_msec
+   {
+      my ( $v ) = @_;
+      return sprintf "%.3f sec", $v / 1000        if $v < 1000*10; $v /= 1000;
+      return sprintf "%.1f sec", $v               if $v < 60;      $v /= 60;
+      return sprintf "%dm%02ds", $v / 60, $v % 60 if $v < 60;      $v /= 60;
+      return sprintf "%dh%02dm", $v / 60, $v % 60;
+   }
+
    sub format_membership
    {
       my ( $membership, $member ) = @_;
@@ -1011,18 +1083,150 @@ package RoomTab {
       $room->send_message( type => "m.emote", body => $text );
    }
 
+   sub cmd_notice
+   {
+      my $self = shift;
+      my ( @args ) = @_;
+
+      my $text = join " ", @args;
+      my $room = $self->{room};
+
+      $room->send_message( type => "m.notice", body => $text );
+   }
+
    sub cmd_image
    {
       my $self = shift;
-      my ( $url ) = @_;
+      my ( $file ) = @_;
 
       my $room = $self->{room};
 
-      $room->send_message( type => "m.image",
-         url  => $url,
-         # TODO: should fetch size, metadata, etc...
-         info => {},
-      );
+      unless( -e $file ) {
+         $self->append_line( "File $file not found!" );
+         return Future->done;
+      }
+
+      my $exifTool = Image::ExifTool->new;
+      $exifTool->ImageInfo( $file );
+
+      my $content_type = $exifTool->GetValue( 'MIMEType' ) //
+         "application/octet-stream";
+
+      $matrix->upload( file => $file, content_type => $content_type )->then( sub {
+         my ( $uri ) = @_;
+
+         $room->send_message( type => "m.image", body => "image attachment",
+            url      => $uri,
+            filename => $file,
+            info     => {
+               size     => $exifTool->GetValue( "FileSize", "ValueConv" ),
+               w        => $exifTool->GetValue( "ImageWidth" ),
+               h        => $exifTool->GetValue( "ImageHeight" ),
+               mimetype => $content_type,
+            },
+         );
+      });
+   }
+
+   sub cmd_audio
+   {
+      my $self = shift;
+      my ( $file ) = @_;
+
+      my $room = $self->{room};
+
+      unless( -e $file ) {
+         $self->append_line( "File $file not found!" );
+         return Future->done;
+      }
+
+      my $exifTool = Image::ExifTool->new;
+      $exifTool->ImageInfo( $file );
+
+      my $content_type = $exifTool->GetValue( "MIMEType" ) //
+         "application/octet-stream";
+
+      $matrix->upload( file => $file, content_type => $content_type )->then( sub {
+         my ( $uri ) = @_;
+
+         $room->send_message( type => "m.audio", body => "audio attachment",
+            url      => $uri,
+            filename => $file,
+            info     => {
+               size     => $exifTool->GetValue( "FileSize", "ValueConv" ),
+               duration => int( $exifTool->GetValue( "Duration", "ValueConv" ) * 1000 ), # msec
+               mimetype => $content_type,
+            },
+         );
+      });
+   }
+
+   sub cmd_video
+   {
+      my $self = shift;
+      my ( $file ) = @_;
+
+      my $room = $self->{room};
+
+      unless( -e $file ) {
+         $self->append_line( "File $file not found!" );
+         return Future->done;
+      }
+
+      my $exifTool = Image::ExifTool->new;
+      $exifTool->ImageInfo($file);
+
+      my $content_type = $exifTool->GetValue( "MIMEType" ) //
+         "application/octet-stream";
+
+      $matrix->upload( file => $file, content_type => $content_type )->then( sub {
+         my ( $uri ) = @_;
+
+         $room->send_message( type => "m.video", body => "video attachment",
+            url      => $uri,
+            filename => $file,
+            info     => {
+               size     => $exifTool->GetValue( "FileSize", "ValueConv" ),
+               duration => int( $exifTool->GetValue( "Duration", "ValueConv" ) * 1000 ), # msec
+               w        => $exifTool->GetValue( "ImageWidth" ),
+               h        => $exifTool->GetValue( "ImageHeight" ),
+               mimetype => $content_type,
+            },
+         );
+      });
+   }
+
+   sub cmd_file
+   {
+      my $self = shift;
+      my ( $file ) = @_;
+
+      my $room = $self->{room};
+
+      unless( -e $file ) {
+         $self->append_line( "File $file not found!" );
+         return Future->done;
+      }
+
+      my $exifTool = Image::ExifTool->new;
+      $exifTool->ImageInfo($file);
+
+      my $content_type = $exifTool->GetValue( "MIMEType" ) //
+         "application/octet-stream";
+
+      $matrix->upload( file => $file, content_type => $content_type )->then( sub {
+         my ( $uri ) = @_;
+
+         $room->send_message( type => "m.file", body => "file attachment",
+            url      => $uri,
+            filename => $file,
+            info     => {
+               size     => $exifTool->GetValue( "FileSize", "ValueConv" ),
+               mimetype => $content_type,
+            },
+         );
+         Future->done;
+      });
    }
 
    sub cmd_leave
@@ -1081,7 +1285,9 @@ package RoomTab {
          $room->set_topic( $topic )
       }
       else {
-         # TODO: Fetch and print the current topic
+         my $room_topic = $room->topic;
+         my $room_name = $room->name;
+         $self->append_line( "Topic for $room_name is: $room_topic" );
          Future->done;
       }
    }
@@ -1097,7 +1303,8 @@ package RoomTab {
          $room->set_name( $name )
       }
       else {
-         # TODO: Fetch and print the current name
+         my $room_name = $room->name;
+         $self->append_line( "Room name is: $room_name" );
          Future->done;
       }
    }
@@ -1122,6 +1329,15 @@ package RoomTab {
       $matrix->add_alias( $alias, $room_id );
    }
 
+   sub cmd_list_aliases
+   {
+      my $self = shift;
+
+      map { $self->append_line( "Room alias: $_" ) } $self->{room}->aliases;
+      Future->done;
+   }
+
+
    sub cmd_delete_alias
    {
       my $self = shift;
