@@ -11,14 +11,14 @@ use warnings;
 use base qw( IO::Async::Notifier );
 IO::Async::Notifier->VERSION( '0.63' ); # adopt_future
 
-our $VERSION = '0.16';
+our $VERSION = '0.17';
 $VERSION = eval $VERSION;
 
 use Carp;
 
 use Future;
 use Future::Utils qw( repeat );
-use JSON qw( encode_json decode_json );
+use JSON::MaybeXS qw( encode_json decode_json );
 
 use Data::Dump 'pp';
 use File::stat;
@@ -31,7 +31,7 @@ struct User => [qw( user_id displayname presence last_active )];
 use Net::Async::Matrix::Room;
 
 use constant PATH_PREFIX => "/_matrix/client/api/v1";
-use constant LONGPOLL_SECONDS => 30;
+use constant LONGPOLL_TIMEOUT => 30;
 
 # This is only needed for the (undocumented) recaptcha bypass feature
 use constant HAVE_DIGEST_HMAC_SHA1 => eval { require Digest::HMAC_SHA1; };
@@ -153,6 +153,11 @@ for events.
 This may be useful in simple send-only cases where the client has no interest
 in receiveing any events, and wishes to reduce the load on the homeserver.
 
+=head2 longpoll_timeout => NUM
+
+Optional. Timeout in seconds for the C</events> longpoll operation. Defaults
+to 30 seconds if not supplied.
+
 =cut
 
 sub _init
@@ -187,6 +192,8 @@ sub _init
    $self->{rooms_by_id} = {};
 
    $self->{path_prefix} = PATH_PREFIX;
+
+   $self->{longpoll_timeout} = LONGPOLL_TIMEOUT;
 }
 
 =head1 METHODS
@@ -201,7 +208,7 @@ sub configure
    my $self = shift;
    my %params = @_;
 
-   foreach (qw( server path_prefix ua SSL enable_events
+   foreach (qw( server path_prefix ua SSL enable_events longpoll_timeout
                 on_log on_unknown_event on_presence on_room_new on_room_del on_invite
                 on_room_member on_room_message )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
@@ -536,6 +543,12 @@ sub _register_with_recaptcha
    };
 }
 
+sub await_synced
+{
+   my $self = shift;
+   return $self->{synced_future} //= $self->loop->new_future;
+}
+
 =head2 $f = $matrix->start
 
 Performs the initial IMSync on the server, and starts the event stream to
@@ -558,6 +571,11 @@ sub start
    defined $self->{access_token} or croak "Cannot ->start without an access token";
 
    return $self->{start_f} ||= do {
+      undef $self->{synced_future};
+      foreach my $room ( values %{ $self->{rooms_by_id} } ) {
+         $room->_reset_for_sync;
+      }
+
       my $f = $self->_do_GET_json( "/initialSync", limit => 0 )
          ->on_fail( sub { undef $self->{start_f} } )
       ->then( sub {
@@ -573,6 +591,7 @@ sub start
                my $room = $self->_get_or_make_room( $room_id );
                $room->_handle_event_initial( $_ ) for @$state;
 
+               $room->await_synced->done;
                $room->maybe_invoke_event( on_synced_state => );
             }
             elsif( $membership eq "invite" ) {
@@ -588,6 +607,7 @@ sub start
             $self->_incoming_event( $_ );
          }
 
+         $self->await_synced->done;
          $self->start_longpoll( start => $sync->{end} ) if $self->{enable_events} // 1;
          Future->done;
       });
@@ -606,7 +626,7 @@ sub stop
 {
    my $self = shift;
 
-   ( delete $self->{start_f} )->cancel;
+   ( delete $self->{start_f} )->cancel if $self->{start_f};
    $self->stop_longpoll;
 }
 
@@ -624,12 +644,12 @@ sub start_longpoll
       my $last_token = $self->{longpoll_last_token};
 
       Future->wait_any(
-         $self->{make_delay}->( LONGPOLL_SECONDS + 5 )
+         $self->{make_delay}->( $self->{longpoll_timeout} + 5 )
             ->else_fail( "Longpoll timed out" ),
 
          $self->_do_GET_json( "/events",
             $last_token ? ( from => $last_token ) : (),
-            timeout => LONGPOLL_SECONDS * 1000, # msec
+            timeout => $self->{longpoll_timeout} * 1000, # msec
          )->then( sub {
             my ( $data ) = @_;
 
@@ -1090,7 +1110,11 @@ sub _handle_event_m_room
    my @subtype_parts;
    while( @type_parts ) {
       if( my $code = $handler->can( "_handle_roomevent_" . join "_", @type_parts, "forward" ) ) {
-         $code->( $handler, @subtype_parts, $event );
+         # Avoid a race condition if a new room event arrives before we finish
+         # the room initial sync
+         $handler->await_synced->on_done( sub {
+            $code->( $handler, @subtype_parts, $event );
+         });
          return;
       }
 
