@@ -11,7 +11,7 @@ use warnings;
 # Not really a Notifier but we like the ->maybe_invoke_event style
 use base qw( IO::Async::Notifier );
 
-our $VERSION = '0.16';
+our $VERSION = '0.17';
 $VERSION = eval $VERSION;
 
 use Carp;
@@ -185,6 +185,30 @@ sub _do_POST_json
    $self->{matrix}->_do_POST_json( "/rooms/$self->{room_id}" . $path, $content );
 }
 
+=head2 $f = $room->await_synced
+
+Returns a L<Future> stored within the room that will complete (with no value)
+once the room initial state sync has been completed. This completes just
+I<before> the C<on_synced_state> event.
+
+=cut
+
+sub _reset_for_sync
+{
+   my $self = shift;
+
+   undef $self->{synced_future};
+
+   undef %{ $self->{state} };
+   undef %{ $self->{members_by_userid} };
+}
+
+sub await_synced
+{
+   my $self = shift;
+   return $self->{synced_future} //= $self->loop->new_future;
+}
+
 sub initial_sync
 {
    my $self = shift;
@@ -192,15 +216,20 @@ sub initial_sync
    # There is not actually an 'initialSync' API for individual rooms, yet. See
    #   https://matrix.org/jira/browse/SYN-55
 
-   $self->{initial_sync} ||= $self->_do_GET_json( "/state" )->then( sub {
-      my ( $events ) = @_;
+   $self->{initial_sync} ||= do {
+      $self->_reset_for_sync;
 
-      $self->_handle_event_initial( $_ ) for @$events;
+      $self->_do_GET_json( "/state" )->then( sub {
+         my ( $events ) = @_;
 
-      $self->maybe_invoke_event( on_synced_state => );
-      Future->done;
-   })
-      ->on_fail( sub { undef $self->{initial_sync} } );
+         $self->_handle_event_initial( $_ ) for @$events;
+
+         $self->await_synced->done( $self );
+         $self->maybe_invoke_event( on_synced_state => );
+         Future->done;
+      })
+         ->on_fail( sub { undef $self->{initial_sync} } );
+   };
 }
 
 sub _handle_event_initial
@@ -902,6 +931,10 @@ Require C<geo_uri>
 
 =back
 
+If an additional argument called C<txn_id> is provided, this is used as the
+transaction ID for the message, which is then sent as a C<PUT> request instead
+of a C<POST>.
+
 =head2 $room->send_message( $text )->get
 
 A convenient shortcut to sending an C<text> message with a body string and
@@ -935,8 +968,14 @@ sub send_message
       $args{$_} or croak "'$type' messages require a '$_' field";
    }
 
-   $self->_do_POST_json( "/send/m.room.message", \%args )
-      ->then_done()
+   if( defined( my $txn_id = $args{txn_id} ) ) {
+      $self->_do_PUT_json( "/send/m.room.message/$txn_id", \%args )
+         ->then_done()
+   }
+   else {
+      $self->_do_POST_json( "/send/m.room.message", \%args )
+         ->then_done()
+   }
 }
 
 =head2 $room->paginate_messages( limit => $n )->get
